#ifndef __LEE_H__
#define __LEE_H__

#include <string>
#include <stdio.h>
#include <iostream>
#include <fstream>
#include <list>
#include <pthread.h>

#include "stm.h"

class GridCell
{
    int m_val;
	
	public:
        // ctor
        GridCell(int val = -1)
            : m_val(val)
        { }

		void set_val(int v) {
			wlpdstm::write_word(&m_val, (void *)v);
		}

		int get_val() {
			return (int)wlpdstm::read_word(&m_val);
		}
};

class Frontier {

	int x, y, z, dw;

	public:
	Frontier(int xx, int yy, int zz, int ddw);
	int getDw();
	void setDw(int ddww);
	int getX();
	void setX(int xx);
	int getY();
	void setY(int yy);
	int getZ();
	void setZ(int zz);
};
    
    
	class Grid {
		
		public:
		int width, height, depth;
		bool releasable;
		GridCell ***grid;
		GridCell ***verifyGrid;
		int debugCount;
		int divisor;
		
		static const int OCC   = 5120;
		static const int VIA   = 6000;
		static const int BVIA  = 6001;
		static const int TRACK = 8192;
		static const int MAX_WEIGHT = 1;
		
		Grid(int gridWidth, int gridHeight, int gridDepth, bool rel);
		void addWeights();
		void occupy(int loX, int loY, int upX, int upY);
		int getWidth();
		int getHeight();
		int getDepth();
		int getPoint(int x, int y, int z);
		void setPoint(int x, int y, int z, int val);
		int getPointNonRelease(int x, int y, int z);
		int getVerifyPoint(int x, int y, int z);
		void setVerifyPoint(int x, int y, int z, int val);
		int getVerifyPointNonRelease(int x, int y, int z);
		void instantiateGrid(GridCell ***g);
		void resetGrid(GridCell ***g);
		bool isValidTrackID(int i);
		void printLayout(bool toFile);
		
	};
	
	class WorkQueue {
		public:
		int x1, y1, x2, y2, netNo;
		long priority;
		double lengthSquared;
		WorkQueue *next;
		
		public:
		WorkQueue();
		WorkQueue(int xs, int ys, int xg, int yg, int nn);
		int getX1();
		int getY1();
		int getX2();
		int getY2();
		int getNetNo();
		void sort();
		void enQueue(int xs, int ys, int xg, int yg, int nn);
		void enQueue(WorkQueue *q);
		WorkQueue * deQueue();
		void toString();
		WorkQueue * getNext();
		void setNext(WorkQueue *q);
		bool equals(WorkQueue *q);
		int listLength();
		bool less(int xx1, int yy1, int xx2, int yy2);
		bool less(WorkQueue *n);
		
		
	};

	struct thread_args_t {
		int id;
	};

#ifdef IRREGULAR_ACCESS_PATTERN
	class ContentionObject {
		public:
		    // ctor
	        ContentionObject(int val = -1)
	            : m_val(val)
	        { }
	
			int get_val() const {
				return (int)wlpdstm::read_word((void *)&m_val);
			}

			void update_val() {
				int val = (int)wlpdstm::read_word(&m_val);
				wlpdstm::write_word((void *)&m_val, (void *)(val + 1));
			}

		private:
			int m_val;
	};

#endif // IRREGULAR_ACCESS_PATTERN


	class Lee {
	
		public:
	
	    static const bool TEST = false;
	    //static const bool DEBUG = true;
	    static const bool DEBUG = false;
	    int GRID_SIZE;
	    static const bool XML_REPORT = false;
	    static const bool VERIFY = true;
	    static const int EMPTY = 0;
	    static const int TEMP_EMPTY = 10000;

	    // note these very useful arrays
		static const int dx[2][4];
		static const int dy[2][4];

		//static int ***private_work_area[lpdstm::MAX_THREADS];

	    int netNo, num_vias, forced_vias, failures, maxTrackLength;
	    Grid *grid;
	    WorkQueue *work, *verifyQueue;
	    pthread_mutex_t queueLock, verifyLock;
	    Lee();
	    Lee(const char* file, bool test, bool debug, bool rel);
	    void parseDataFile(const char *file);
	    void fakeTestData();
	    int readInt(std::string *line);
	    WorkQueue* getNextTrack();
	    void addTrackForVerification(WorkQueue *q);
		void removeTrackFromVerification(WorkQueue *q);
		bool ok(int x, int y);
		int deviation(int x1, int y1, int x2, int y2);
		bool expandFromTo(int x, int y, int xGoal, int yGoal,
		int num, int ***tempg);
		bool backtrackFrom(int xGoal, int yGoal, int xStart,
		int yStart, int trackNo, int ***tempg);
		int trackLength(int x1, int y1, int x2, int y2);
		bool pathFromOtherSide(int ***g, int X, int Y, int Z);
		bool connect(WorkQueue *q, int ***tempg); 
		bool layNextTrack(WorkQueue *q, int ***tempg);

#ifdef IRREGULAR_ACCESS_PATTERN
		ContentionObject contention_object;

		unsigned read_contention_object() {
			return contention_object.get_val();
		}

		void update_contention_object() {
			contention_object.update_val();
		}
#endif // IRREGULAR_ACCESS_PATTERN

	};
	
	
#endif /*__LEE_H__*/
